#include "stdlib.h"
#include "string.h"
#include "sys/types.h"

#define __need_ptrdiff_t
#define __need_NULL
#include <stddef.h>

#ifndef TEST
# define sbrk crystal_sbrk
#else
# include <stdio.h>
# define fdprintf(fd, fmt...)	fprintf (stderr, fmt)
#endif

extern void *sbrk (ptrdiff_t);

typedef struct header
  {
    struct header *next;
    size_t size;
  }
HEADER;

static HEADER *free_list;

void
free (void *vp)
{
  HEADER *p;
  HEADER *being_freed = vp;
  being_freed--;

  fdprintf (2, "Entering free (%p)\n", vp);
  for (p = (HEADER *) &free_list;
       p->next && p->next <= being_freed;
       p = p->next)
    ;
  if (free_list != p->next && (char *) p + p->size == (char *) being_freed)
    {
      p->size += being_freed->size;
      if (p->next
	  &&
	  (char *) being_freed + being_freed->size == (char *) p->next)
	{
	  p->size += p->next->size;
	  p->next = p->next->next;
	}
    }
  else if (p->next
	   &&
	   (char *) being_freed + being_freed->size == (char *) p->next)
    {
      being_freed->next = p->next->next;
      being_freed->size += p->next->size;
      p->next = being_freed;
    }
  else
    {
      being_freed->next = p->next;
      p->next = being_freed;
    }
  fdprintf (2, "Exiting free (%p)\n", vp);
}

void *
malloc (size_t size)
{
  size_t incr = ((((size + sizeof (HEADER) - 1) / sizeof (HEADER)) + 1)
		 * sizeof (HEADER));
#if 0
  HEADER *p = sbrk (incr);
  void *ap;

  ap = NULL;
  if (p != (HEADER *)-1)
    {
      p->size = incr;
      ap = p + 1;
    }
  fdprintf (2, "malloc (%u) = %p\n", size, ap);
  return ap;
#else
  HEADER *p, *q;

  fdprintf (2, "Entering malloc (%u), incr = %u\n", size, incr);

  for (p = (HEADER *) &free_list; p->next; p = p->next)
    if (p->next->size >= incr)
      {
	q = p->next;
	if (q->size - incr >= sizeof (HEADER) * 2)
	  {
	    p->next = (HEADER *) ((char *) q + incr);
	    p->next->next = q->next;
	    p->next->size = q->size - incr;
	    q->size = incr;
	  }
	else
	  p->next = q->next;
	fdprintf (2, "Exiting malloc (%u) = %p\n", size, q + 1);
	return q + 1;
      }

  fdprintf (2, "malloc: Free area not found, try to sbrk()\n");
  if (free_list != p->next && (char *) sbrk (0) != (char *) p + p->size)
    {
      if (sbrk (incr - p->size) == (void *)-1)
	goto oom;
      p->size = incr;

      fdprintf (2, "Exiting malloc (%u) = %p\n", size, p + 1);
      return p + 1;
    }

  fdprintf (2, "Trying to sbrk(%u)\n", incr);
  if ((q = sbrk (incr)) == (void *)-1)
    goto oom;
  q->next = NULL;
  q->size = incr;
  fdprintf (2, "Exiting malloc (%u) = %p\n", size, q + 1);
  return q + 1;

oom:
  fdprintf (2, "malloc: Cannot allocate %u bytes memory block\n", size);
  return NULL;
#endif  
}

void *
realloc (void *old, size_t newsize)
{
  HEADER *p, *q, *r;
  void *new;
  size_t incr;
  size_t avail;

  if (old == NULL)
    return malloc (newsize);
  if (!newsize)
    {
      free (old);
      return NULL;
    }

  p = old;
  --p;
  incr = ((((newsize + sizeof (HEADER) - 1) / sizeof (HEADER)) + 1)
	  * sizeof (HEADER));
  if (incr > p->size)
    {
      for (q = r = (HEADER *) &free_list; q->next; r = q, q = q->next)
	if (p < q->next)
	  break;
  
      avail = p->size;
      if ((char *) p + p->size == (char *) q->next)
	{
	  avail += q->next->size;
	  if (avail >= incr)
	    {
	      p->size = avail;
	      q->next = q->next->next;
	      goto do_split;
	    }
	}

      if (r != (HEADER *) &free_list && (char *) q + q->size == (char *) p)
	{
	  avail += r->size;
	  if (avail >= incr)
	    {
	      if ((char *) p + p->size == (char *) q->next)
		r->next = q->next->next;
	      else
		r->next = q->next;
	      q->size = avail;
	      memcpy (q + 1, p + 1, p->size - sizeof (HEADER));
	      p = q;
	      goto do_split;
	    }
	}

      if (!q->next && (char *) sbrk (0) == (char *) p + p->size)
	{
	  if (sbrk (incr - p->size) == (void *)-1)
	    goto oom;
	  p->size = incr;
	  return p + 1;
	}

      if ((new = malloc (newsize)) != NULL)
	{
	  memcpy (new, p + 1, p->size);
	  free (p + 1);
	}

      return new;
    }

do_split:
  if (p->size - incr >= 2 * sizeof (HEADER))
    {
      q = (HEADER *) ((char *) p + incr);
      q->size = p->size - incr;
      p->size = incr;
      free (q + 1);
    }

  return p + 1;

oom:
  fdprintf (2, "realloc: cannot re-allocate %d bytes block\n", newsize);
  return NULL;
}

#ifdef TEST
# include <stdio.h>
# include <stdlib.h>

static char heap[1 << 20];

int
main (int argc, char *argv[])
{
#if 0
  char buf[256];

  printf ("Heap: %p - %p\n", heap, heap + sizeof (heap) - 1);

  while (fputs ("> ", stdout), fgets (buf, sizeof (buf), stdin) != NULL)
    {
      HEADER *p;
      size_t size;

      if (sscanf (buf, "%p %u", &p, &size) != 2)
	continue;
      p->size = size;
      free (p + 1);

      for (p = free_list; p; p = p->next)
	printf ("%#08x + %8u\n", (unsigned int) p, p->size);
    }

  return 0;
#else
  void *p1, *p2, *p3, *p4;
  p1 = malloc (200);
  p2 = malloc (100);
  p3 = realloc (p1, 300);
  free (p2);
  p4 = realloc (p3, 150);
  free (p4);
  p1 = malloc (50);
  p2 = malloc (50);
  p3 = realloc (p1, 20);
  p4 = malloc (10);
  printf ("p1 = %p, p2 = %p, p3 = %p, p4 = %p\n", p1, p2, p3, p4);
#endif
}
#endif
